/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.calculation;

import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.Modifier;
import org.intoJ.inferType.exception.InternalCalculationException;
import org.intoJ.inferType.model.collection.DESet_Signature;
import org.intoJ.inferType.model.collection.HierarchySet;
import org.intoJ.inferType.model.collection.TypeSet;
import org.intoJ.inferType.model.model.SingleDeclarationElement;
import org.intoJ.inferType.model.model.TypeElement;
import org.intoJ.inferType.util.Util_Conversion;

/**
 * This class creates a fully resolved class and interface hierarchy, starting
 * from an IType and going upwards. The resulting "TypeElements" contain ALL
 * fields and methods of the original types (including, in contrast to the
 * original ITypeBindings, the inherited ones and, in contrast to the original
 * ITypes, with fully resolved names in parameter names etc).
 * 
 * Note that the members are compared via SIGNATURE, not via KEY. This means a
 * type will contain an inherited and overridden member only ONCE. It is
 * unspecified which concrete object is being stored (the inherited or
 * overridden one) - do not use the keys anymore!
 * 
 * 
 * @version: $Date: 2007/07/28 12:12:55 $Author: kegel $Revision: 1.1 $
 * @author Philip Mayer
 */
public class HierarchyBuilder {

	/**
	 * Builds the supertype hierarchy of the given IType.
	 * 
	 * The hierarchy is returned as a TypeSet, which is a list of TypeElements.
	 * Each TypeElement in that list is a supertype - either class or interface -
	 * of the given IType, which is being returned as a TypeElement, too.
	 * 
	 * Each TypeElement in the list as well as the root will have their methods
	 * and fields resolved, that is root will have any methods which can be
	 * invoked on it. These include:
	 * 
	 * 1) all inherited methods and fields (excluding private ones) 
	 * 2) all declared methods and fields (all of them, including privates) 
	 * 3) No constructors 
	 * 4) Static fields and methods if part of 1) or 2) 
	 * 5) Every field or method, no matter if overridden, will only appear ONCE.
	 * 
	 * Every "top level" type will have a "java.lang.Object" super TypeElement, and 
	 * it is is always the same as per object identity.
	 * 
	 * @param type
	 * 
	 * @return HierarchySet
	 * @throws InternalCalculationException
	 */
	public static HierarchySet buildSuperTypeHierarchy(IType type) throws InternalCalculationException {

		ITypeBinding b = Util_Conversion.getTypeBinding(type);
		return buildSuperTypeHierarchy(b);
	}

	/**
	 * Returns a HierarchySet for an ITypeBinding.
	 * 
	 * @param typeBinding
	 * @return HierarchySet
	 * @throws InternalCalculationException
	 * @see buildSuperTypeHierarchy()
	 * 
	 */
	public static HierarchySet buildSuperTypeHierarchy(ITypeBinding typeBinding) {

		TypeSet col = new TypeSet();
		TypeElement root = getFullTypeElement(typeBinding, col, true);

		HierarchySet set = new HierarchySet(root, col);
		return set;
	}

	/**
	 *
	 * Creates a complete type element.
	 * 
	 * @param binding Current binding to be investigated on (class or interface)
	 * @param collectingParameter (list of types, current one is to be added)
	 * @param root true if current type is root (needed because private members
	 *          of root are to be added).
	 * @return
	 */
	private static TypeElement getFullTypeElement(ITypeBinding binding, TypeSet collectingParameter, boolean root) {

		/**
		 * The supertype of a class is always Object if no other superclasses are present. In these cases,
		 * the binding exists and points to Object. Interfaces posess a "null" top binding.
		 */
		if ((binding == null) || (binding.getQualifiedName().equals("java.lang.Object"))) {

			TypeElement t = new TypeElement("java.lang.Object", false, false);
			if (collectingParameter.getTypeElement("java.lang.Object") == null)
				collectingParameter.addType(t);
			return t;
		}

		// Get members:
		DESet_Signature membersOfTheCurrentTypeOnly = new DESet_Signature();

		// Get the fields
		IVariableBinding[] ifields = binding.getDeclaredFields();
		for (int i = 0; i < ifields.length; i++) {
			if ((root) || (!Modifier.isPrivate(ifields[i].getModifiers()))) {
				membersOfTheCurrentTypeOnly.addDeclaration(new SingleDeclarationElement(ifields[i]));
			}
		}

		// Get the methods
		IMethodBinding[] imethods = binding.getDeclaredMethods();
		for (int i = 0; i < imethods.length; i++) {
			if ((!imethods[i].isConstructor()) && ((root) || (!Modifier.isPrivate(imethods[i].getModifiers()))))
				membersOfTheCurrentTypeOnly.addDeclaration(new SingleDeclarationElement(imethods[i]));
		}

		// Build it

		TypeElement returner = new TypeElement(binding);
		returner.addDeclaredMembers(membersOfTheCurrentTypeOnly);

		// Check superclasses?
		TypeElement superclass = getFullTypeElement(binding.getSuperclass(), collectingParameter, false); // RECURSION
		returner.addInheritedMembers(superclass.getAllMembers());
		returner.addDeclaredSupertype(superclass);
		returner.addInheritedSupertype(superclass.getAllSuperTypes());

		// Check interfaces?
		ITypeBinding[] interfaces = binding.getInterfaces();
		for (int i = 0; i < interfaces.length; i++) {
			TypeElement currentIf = getFullTypeElement(interfaces[i], collectingParameter, false); // RECURSION
			returner.addInheritedMembers(currentIf.getAllMembers());
			returner.addDeclaredSupertype(currentIf);
			returner.addInheritedSupertype(currentIf.getAllSuperTypes());
		}

		collectingParameter.addType(returner);
		return returner;
	}

}