/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.util;

import java.util.Collection;
import java.util.HashSet;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IInitializer;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.Signature;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.search.IJavaSearchConstants;
import org.eclipse.jdt.core.search.IJavaSearchScope;
import org.eclipse.jdt.core.search.SearchEngine;
import org.eclipse.jdt.core.search.SearchParticipant;
import org.eclipse.jdt.core.search.SearchPattern;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.PartInitException;
import org.intoJ.inferType.model.model.CompositeDeclarationElement;
import org.intoJ.inferType.model.model.DeclarationElement;
import org.intoJ.inferType.model.model.SingleDeclarationElement;

/**
 * 
 * This class provides functions for working with ITypes and ITypeBindings.
 * 
 * @version: $Date: 2007/07/28 12:12:55 $Author: kegel $Revision: 1.1 $
 * @author Philip Mayer
 */
public class Util_ITypes {

	/**
	 * Opens the IJavaElement in the editor.
	 * 
	 * @param element IJavaElement to open.
	 */
	public static void openInEditor(final IJavaElement element) {

		Display d = Display.getDefault();
		d.asyncExec(new Runnable() {

			public void run() {

				try {
					if (element != null) {
						ICompilationUnit cu = (ICompilationUnit) element.getAncestor(IJavaElement.COMPILATION_UNIT);
						IEditorPart javaEditor = JavaUI.openInEditor(cu);
						JavaUI.revealInEditor(javaEditor, element);
					}

				} catch (PartInitException e) {
					ILog.logError("EditorUtility.openInEditor", e);
				} catch (JavaModelException e) {
					ILog.logError("EditorUtility.openInEditor", e);
				}
			}
		});
	}

	/**
	 * 
	 * Returns the IType corresponding to the given full qualified type name,
	 * which must reside within the compilation unit. Note that inner (but not
	 * local!) types are found as well.
	 * 
	 * The full qualified name must be separated by dots (including the inner
	 * type names - for example: org.intoJ.core.ClassA.ClassB).
	 * 
	 * 
	 * @param fqTypeNameDotSeparated
	 * @param compilationUnit
	 * @return IType if found, null if not
	 */
	public static IType locateITypeInsideSourceFile(String fqTypeNameDotSeparated, IJavaElement compilationUnit) {

		if (compilationUnit instanceof ICompilationUnit) {

			// Load all types, including inner types (but first without local
			// types, declared in methods)

			IType[] theTypes = null;
			try {
				theTypes = ((ICompilationUnit) compilationUnit).getAllTypes();
				int found = -1;
				for (int i = 0; i < theTypes.length; i++) {
					//BUG: typeBinding.getQualifiedName() marks inner classes
					// with '.', type.getFullyQualifiedName() with '$'.
					if (theTypes[i].getFullyQualifiedName().replace('$', '.').equals(fqTypeNameDotSeparated)) { // Use the type's
						// name here!
						found = i;
						break;
					}
				}
				if ((found > -1) && (theTypes[found].exists()))
					return theTypes[found];

				// Type was not found. It definitely exists, so try to find a
				// local type. Parsing!
				for (int i = 0; i < theTypes.length; i++) {
					IInitializer[] inits = theTypes[i].getInitializers();
					IMethod[] methods = theTypes[i].getMethods();
					for (int j = 0; j < inits.length; j++) {
						IType tryit = inits[j].getType(Signature.getSimpleName(fqTypeNameDotSeparated), 1);
						if (tryit.exists())
							return tryit;
					}
					for (int k = 0; k < methods.length; k++) {
						IType tryit = methods[k].getType(Signature.getSimpleName(fqTypeNameDotSeparated), 1);
						if (tryit.exists())
							return tryit;
					}
				}

			} catch (JavaModelException e2) {
				e2.printStackTrace();
			}
		}
		return null;
	}

	/**
	 * This method finds the ICompilationUnits with references to the given
	 * declaration element.
	 * @param declarationElement
	 * @param scope the SearchScope
	 * 
	 * @return collection of ICompilationUnits
	 * @throws JavaModelException
	 */
	public static Collection getCompilationUnits(DeclarationElement declarationElement, IJavaSearchScope scope) throws JavaModelException {

		if (declarationElement.getKind() != DeclarationElement.METHOD && declarationElement.getKind() != DeclarationElement.FIELD
				&& declarationElement.getKind() != DeclarationElement.LOCALVARIABLE && declarationElement.getKind() != DeclarationElement.PARAMETER)
			return null;

		Collection compilationUnits = new HashSet();

		/**
		 * We need to find the compilation units "associated" with the given DeclarationElement - that is, in which
		 * the element occurs. For methods and fields, this may be any class in the system (if they are public); parameters
		 * and locals can only occur within their defining type.
		 * 
		 * * Methods and fields are searched using their element names, like "field1" and "method1". 
		 * * Params and locals are searched using the fully qualified name of the types they are defined in. 
		 * 
		 */

		int searchFor = 0;
		int forWhat = 0;

		SearchResultCollector compilationUnitCollector = new SearchResultCollector();
		SearchPattern searchPattern;

		if ((declarationElement.getKind() == IBinding.METHOD || declarationElement.getKind() == DeclarationElement.FIELD)) {

			if (declarationElement.getKind() == IBinding.METHOD)
				searchFor = IJavaSearchConstants.METHOD;
			else
				searchFor = IJavaSearchConstants.FIELD;

			forWhat = IJavaSearchConstants.ALL_OCCURRENCES;

			if (declarationElement instanceof SingleDeclarationElement) {
				searchPattern = SearchPattern.createPattern(declarationElement.getSimpleName(), searchFor, forWhat, SearchPattern.R_EXACT_MATCH
						| SearchPattern.R_CASE_SENSITIVE);
			} else {
				IBinding[] bindings = ((CompositeDeclarationElement) declarationElement).getBindings();

				searchPattern = SearchPattern.createPattern(bindings[0].getName(), searchFor, forWhat, SearchPattern.R_EXACT_MATCH
						| SearchPattern.R_CASE_SENSITIVE);
				for (int i = 1; i < bindings.length; i++) {
					SearchPattern tmpPattern = SearchPattern.createPattern(bindings[i].getName(), searchFor, forWhat, SearchPattern.R_EXACT_MATCH
							| SearchPattern.R_CASE_SENSITIVE);
					searchPattern = SearchPattern.createOrPattern(searchPattern, tmpPattern);
				}
			}
		} else { // Local or param: Search for the type (note that we need to SEARCH for it. javaProject.findType() does not find inner types.
			searchFor = IJavaSearchConstants.TYPE;
			forWhat = IJavaSearchConstants.DECLARATIONS;

			if (declarationElement instanceof SingleDeclarationElement) {
				//FIXME: Anonymous Inner Classes have no name?!
				
				searchPattern = SearchPattern.createPattern(((SingleDeclarationElement) declarationElement).getDeclaringClass().getQualifiedName(), searchFor,
						forWhat, SearchPattern.R_EXACT_MATCH | SearchPattern.R_CASE_SENSITIVE);
			} else {
				ITypeBinding[] bindings = ((CompositeDeclarationElement) declarationElement).getDeclaringClasses();

				searchPattern = SearchPattern.createPattern(bindings[0].getQualifiedName(), searchFor, forWhat, SearchPattern.R_EXACT_MATCH
						| SearchPattern.R_CASE_SENSITIVE);
				for (int i = 1; i < bindings.length; i++) {
					SearchPattern tmpPattern = SearchPattern.createPattern(bindings[i].getQualifiedName(), searchFor, forWhat, SearchPattern.R_EXACT_MATCH
							| SearchPattern.R_CASE_SENSITIVE);
					searchPattern = SearchPattern.createOrPattern(searchPattern, tmpPattern);
				}
			}

		}

		SearchParticipant searchParticipant = SearchEngine.getDefaultSearchParticipant();
		SearchEngine searchEngine = new SearchEngine();
		try {
			//FIXME searchPattern might be "", i.e. anonymous inner classes
			if( !(searchPattern == null) )
			searchEngine.search(searchPattern, new SearchParticipant[] { searchParticipant }, scope, compilationUnitCollector, null);
		} catch (CoreException e) {
			throw new JavaModelException(e);
		}
		compilationUnits = compilationUnitCollector.getCompilationUnits();
		return compilationUnits;
	}

	/**
	 * This method parses a given ICompilationUnit.
	 * 
	 * @param icu
	 *            the ICompilationUnit
	 * @param resolveBindings
	 *            should compiler provide binding information
	 * 
	 * @return the corresponding CompilationUnit
	 */
	public static CompilationUnit parseCompilationUnit(ICompilationUnit icu, boolean resolveBindings) {

		ASTParser parser = ASTParser.newParser(AST.JLS3);
		parser.setSource(icu);
		parser.setResolveBindings(resolveBindings);
		CompilationUnit currentCU = (CompilationUnit) parser.createAST(null);
		return currentCU;
	}

}