/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.multirefac.gathering;

import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.ChildListPropertyDescriptor;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.PrimitiveType;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationExpression;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.intoJ.inferType.calculation.VariableDeclarationExpressionStatement;
import org.intoJ.inferType.model.collection.DESet_Key;
import org.intoJ.inferType.model.model.SingleDeclarationElement;
import org.intoJ.inferType.util.ILog;

/**
 * 
 * This visitor finds all variables and methods declared with the given set of
 * ITypes. The references are put into a DeclarationSet (Collecting parameter
 * pattern).
 * 
 * Warning: Overridden fields and methods and parameters will be inserted as
 * separate declaration elements.
 * 
 * @version: $Date: 2007/07/28 12:12:55 $Author: kegel $Revision: 1.1 $
 * @author Andreas Mei&szlig;ner
 */
public class DeclarationElementVisitor extends ASTVisitor {

	private DESet_Key declarationElementsMap;
	private Set types;

	/**
	 * Constructor for DeclarationElementVisitor.
	 * @param types Set
	 * @param deMap DESet_Key
	 */
	public DeclarationElementVisitor(Set types, DESet_Key deMap) {

		this.declarationElementsMap = deMap;
		this.types = types;
	}

	/**
	 * Visiting methods. INCLUDES Constructors.
	 * 
	 * @param node MethodDeclaration
	 * @return boolean
	 */
	public boolean visit(MethodDeclaration node) {

		if (node == null)
			return true;
		ITypeBinding typeBinding;
		if (node.getReturnType2() == null) { // Constructor
			typeBinding = node.getAST().newPrimitiveType(PrimitiveType.VOID).resolveBinding();
		} else {
			typeBinding = node.getReturnType2().resolveBinding();
		}
		if (typeBinding == null)
			return true;

		ILog.log(ILog.LOG_AST, 4, "Visiting METHOD DECLARATION / node: " + node.toString());

		if (isPartOfSearch(typeBinding)) {
			IMethodBinding currentMethodBinding = node.resolveBinding();
			if (currentMethodBinding == null)
				return true;

			ILog.log(ILog.LOG_AST, 4, "Adding MethodDeclaration " + currentMethodBinding + " to map.");

			declarationElementsMap.addDeclaration(new SingleDeclarationElement(currentMethodBinding, node.getReturnType2().getStartPosition()));
		}

		// Parameters. They need to be found HERE, because we need the index of
		// the parameter in the list.

		List params = node.parameters();
		int indexInMethod = 0;
		for (Iterator iter = params.iterator(); iter.hasNext();) {
			SingleVariableDeclaration param = (SingleVariableDeclaration) iter.next();

			ITypeBinding paramTypeBinding = param.getType().resolveBinding();
			if (paramTypeBinding == null)
				return true;

			IMethodBinding currentMethodBinding = node.resolveBinding();
			if (currentMethodBinding == null)
				return true;

			if (isPartOfSearch(paramTypeBinding)) {
				IVariableBinding currentVariableBinding = param.resolveBinding();

				// Need to check again(!) due to array issues!
				if (isPartOfSearch(currentVariableBinding.getType())) {

					if (currentVariableBinding != null) {
						ITypeBinding declaringType = getDeclaringType(param);
						ILog.log(ILog.LOG_AST, 4, "Adding Parameter " + currentVariableBinding + " to map.");
						declarationElementsMap.addDeclaration(new SingleDeclarationElement(currentVariableBinding, param.getType().getStartPosition(), declaringType,
								indexInMethod, currentMethodBinding));
					}
				}
			}
			indexInMethod++;
		}

		return true;
	}

	/**
	 * For statement variables.
	 * 
	 * VariableDeclarationExpression ( <Expression <ASTNode) is just for
	 * variables in for loops 
	 * 
	 * VariableDeclarationStatement ( <Statement
	 * <ASTNode) is for all the other variables.
	 * 
	 * Note that, as of Eclipse 3.0 Final, VariableDeclarationExpression and
	 * VariableDeclarationStatement share exactly the same methods for accessing
	 * the fragments (.fragments(), .getType()). However, there is no common
	 * supertype except ASTNode, which does not contain these methods.
	 * 
	 * Copy&Paste will therefore work. However, it seemed better do extract a
	 * common interface and handle it via delegation.
	 * 
	 */
	public boolean visit(VariableDeclarationExpression node) {

		return extractFragments(new VariableDeclarationExpressionStatement(node));
	}

	/**
	 * Visiting local variables
	 * @param node VariableDeclarationStatement
	 * @return boolean
	 */
	public boolean visit(VariableDeclarationStatement node) {

		return extractFragments(new VariableDeclarationExpressionStatement(node));
	}

	/**
	 * Visiting fields.
	 * @param node FieldDeclaration
	 * @return boolean
	 */
	public boolean visit(FieldDeclaration node) {

		if (node == null)
			return true;
		List declaredFields = node.fragments();
		ITypeBinding typeBinding = node.getType().resolveBinding();

		if (typeBinding == null || declaredFields == null)
			return true;

		ILog.log(ILog.LOG_AST, 4, "Visiting FIELD DECLARATION  / node: " + node.toString());

		if (isPartOfSearch(typeBinding)) {
			Iterator fieldsIterator = declaredFields.iterator();
			while (fieldsIterator.hasNext()) {
				VariableDeclarationFragment currentFragment = (VariableDeclarationFragment) fieldsIterator.next();
				IVariableBinding currentVariableBinding = currentFragment.resolveBinding();
				// The variable itself may have additional array information.
				if ((currentVariableBinding != null)) { //ARRAY && (currentFragment.getExtraDimensions() == 0)) {
					ILog.log(ILog.LOG_AST, 4, "Adding FieldDeclaration " + currentVariableBinding + " to map.");
					declarationElementsMap.addDeclaration(new SingleDeclarationElement(currentVariableBinding, node.getType().getStartPosition()));
				}
			}
		}

		return true;
	}

	/**
	 * Visiting formal parameter lists and catch clauses. Note that we won't look for parameters in this method,
	 * but rather in visit(MethodDeclaration), as we need the index number of the parameters which is not
	 * available here.
	 * @param node SingleVariableDeclaration
	 * @return boolean
	 */
	public boolean visit(SingleVariableDeclaration node) {

		if (node == null)
			return true;

		if (node.getLocationInParent() instanceof ChildListPropertyDescriptor) {
			ChildListPropertyDescriptor clpd = (ChildListPropertyDescriptor) node.getLocationInParent();
			if ((clpd != null) && (clpd.getId().equals("parameters")))
				return true; // parameters are found elsewhere.
		}

		ITypeBinding typeBinding = node.getType().resolveBinding();
		if (typeBinding == null)
			return true;

		ILog.log(ILog.LOG_AST, 4, "Visiting SINGLE VARIABLE DECLARATION " + " / node: " + node.toString());
        
		if (isPartOfSearch(typeBinding)) {
			IVariableBinding currentVariableBinding = node.resolveBinding();

			if (currentVariableBinding != null) {

				ITypeBinding declaringType = getDeclaringType(node);
				ILog.log(ILog.LOG_AST, 4, "Adding SingleVariableDeclaration " + currentVariableBinding + " to map.");
				declarationElementsMap.addDeclaration(new SingleDeclarationElement(currentVariableBinding, node.getType().getStartPosition(), declaringType));
			}
		}

		return true;
	}

	private boolean extractFragments(VariableDeclarationExpressionStatement node) {

		if (node == null)
			return true;
		List declaredVariables = node.fragments();
		ITypeBinding typeBinding = node.getType().resolveBinding();
		if (typeBinding == null || declaredVariables == null)
			return true;

		ILog.log(ILog.LOG_AST, 4, "Visiting VARIABLE DECLARATION STATEMENT " + " / node: " + node.toString());

		if (isPartOfSearch(typeBinding)) {
			Iterator iter = declaredVariables.iterator();
			while (iter.hasNext()) {
				VariableDeclarationFragment currentFragment = (VariableDeclarationFragment) iter.next();

				IVariableBinding currentVariableBinding = currentFragment.resolveBinding();
				if (currentVariableBinding == null)
					return true;

				ITypeBinding declaringType = getDeclaringType(node.getASTNode());
				ILog.log(ILog.LOG_AST, 4, "Adding VariableDeclaration " + currentVariableBinding + " to map.");
				declarationElementsMap
						.addDeclaration(new SingleDeclarationElement(currentVariableBinding, node.getType().getStartPosition(), declaringType));
			}
		}

		return true;
	}

    private boolean isPartOfSearch(ITypeBinding typeBinding) {
        if (typeBinding.isWildcardType()) {
            typeBinding = typeBinding.getBound();
            if (typeBinding == null) {
                return false;
            }
        }
        if (typeBinding.isArray()) {
            typeBinding = typeBinding.getElementType();
        }
        if (types.contains(typeBinding.getJavaElement())) {
            return true;
        } else if (typeBinding.isParameterizedType()) {
            for (ITypeBinding argument : typeBinding.getTypeArguments()) {
                if (isPartOfSearch(argument)) {
                    return true;
                }
            }
        }
        return false;
    }

	private ITypeBinding getDeclaringType(ASTNode node) {

		ASTNode cnode = node;
		while (cnode != null) {
			if (cnode instanceof TypeDeclaration)
				return ((TypeDeclaration) cnode).resolveBinding();
			cnode = cnode.getParent();
		}

		return null;
	}
}