/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3;

import java.lang.reflect.InvocationTargetException;
import java.util.List;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.ltk.core.refactoring.Change;
import org.eclipse.ltk.core.refactoring.PerformChangeOperation;
import org.eclipse.ltk.core.refactoring.RefactoringCore;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.eclipse.ltk.ui.refactoring.RefactoringUI;
import org.eclipse.swt.widgets.Display;
import org.intoJ.inferType3.internal.constraintModel.util.SelectedElement;
import org.intoJ.inferType3.internal.constraintModel.util.SelectedElementVisitor;

/**
 * This runnable can be used to start Infer Type from a different plugin without worrying
 * about the eclipse refactoring framework.
 * 
 * @author Hannes Kegel
 *
 */
public class ExternalInferTypeRunnable implements IRunnableWithProgress {
    
    private InferTypeRefactoring refactoring;
    private final boolean changeTypeArgs;
    private final boolean noAbstractClasses;
    private final boolean checkOnly;
    
    private RefactoringStatus refactoringStatus;

    
    /**
     * The general constructor with bindings and type offsets.
     * 
     * @param binding the binding to be refactored
     * @param declaringClass its declaring type
     * @param typeStartPosition the start position of the type (not of the identifier)
     * @param typeVariableKeys if the selection is a type argument of the binding, this list
     * holds all binding keys to the selected (nested) type. Otherwise null.
     * @param clearCache whether to clear the cache (should be set when used the first time)
     * @param changeTypeArgs if type arguments should be changed if necessary
     * @param noAbstractClasses if set, no abstract classes will be generated for case 3
     * @param checkOnly if set, the refactoring will only check if the selection can be generalized. Call
     * {@link #run(IProgressMonitor)} and afterwards {@link #wasApplicable()}.
     */
    public ExternalInferTypeRunnable(IBinding binding, ITypeBinding declaringClass, int typeStartPosition, List<String> typeVariableKeys, boolean clearCache, boolean changeTypeArgs, boolean noAbstractClasses, boolean checkOnly) {
        this.changeTypeArgs = changeTypeArgs;
        this.noAbstractClasses = noAbstractClasses;
        this.checkOnly = checkOnly;
        SelectedElement selectedElement = new SelectedElement(binding, declaringClass, typeStartPosition, typeVariableKeys);
        selectedElement.setCompilationUnit((ICompilationUnit)binding.getJavaElement().getAncestor(IJavaElement.COMPILATION_UNIT));        
        
        refactoring = new InferTypeRefactoring(selectedElement, clearCache);
    }
    
    /**
     *  A constructor with name offsets only. 
     * 
     * @param icu the compilation unit where the selection was done
     * @param nameStartPosition the start position of the name (not of the type)
     * @param nameLength the length of the name (not of the type)
     * @param typeVariableKeys if the selection is a type argument of the binding, this list
     * holds all binding keys to the selected (nested) type. Otherwise null.
     * @param clearCache whether to clear the cache (should be set when used the first time)
     * @param changeTypeArgs if type arguments should be changed if necessary
     * @param noAbstractClasses if set, no abstract classes will be generated for case 3
     * @param checkOnly if set, the refactoring will only check if the selection can be generalized. Call
     * {@link #run(IProgressMonitor)} and afterwards {@link #wasApplicable()}.
     */
    public ExternalInferTypeRunnable(ICompilationUnit icu, int nameStartPosition, int nameLength, List<String> typeVariableKeys, boolean clearCache, boolean changeTypeArgs, boolean noAbstractClasses, boolean checkOnly) {
        this.changeTypeArgs = changeTypeArgs;
        this.noAbstractClasses = noAbstractClasses;
        this.checkOnly = checkOnly;
                
        SelectedElement selectedElement = getSelectedElement(icu, nameStartPosition, nameLength);                
        refactoring = new InferTypeRefactoring(selectedElement, clearCache);
    }    
    
    
    /* (non-Javadoc)
     * @see org.eclipse.jface.operation.IRunnableWithProgress#run(org.eclipse.core.runtime.IProgressMonitor)
     */
    public void run(final IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
        try {
            monitor.beginTask("", 200);
            refactoringStatus = refactoring.checkInitialConditions(new SubProgressMonitor(monitor, 170));
            if (!checkOnly && !refactoringStatus.hasFatalError()) {
                refactoring.setChangeTypeArguments(changeTypeArgs);
                refactoring.setNoAbstractClasses(noAbstractClasses);
                refactoringStatus.merge(refactoring.checkFinalConditions(new SubProgressMonitor(monitor, 10)));
                if (!refactoringStatus.hasFatalError()) {
                    final Change change = refactoring.createChange(new SubProgressMonitor(monitor, 5));
                    if (change != null) {
                        change.initializeValidationData(new SubProgressMonitor(monitor, 5));                        
                        final PerformChangeOperation operation = RefactoringUI.createUIAwareChangeOperation(change);
                        operation.setUndoManager(RefactoringCore.getUndoManager(), "Infer Type");
                        final CoreException[] exHolder = new CoreException[1];
                        Display.getDefault().syncExec(
                                new Runnable() {
                                  public void run(){
                                      try {
                                        operation.run(new SubProgressMonitor(monitor, 10));
                                    } catch (CoreException e) {
                                        exHolder[0] = e;
                                    }
                                  }
                                });
                        if (exHolder[0] != null) {
                            throw exHolder[0];
                        }
                    }
                }

            }
        } catch (OperationCanceledException e) {
            throw e;
        } catch (Throwable e) {
            e.printStackTrace();
            throw new InvocationTargetException(e);
        } finally {
            monitor.done();
        }
    }

    /**
     * Retrieves the error message if there was an error. Otherwise returns null.
     * 
     * @return the message
     */
    public String getErrorMessage() {
        return refactoringStatus.getMessageMatchingSeverity(RefactoringStatus.FATAL);
    }

    /**
     * Returns the number of changed type occurences.
     * 
     * @return the number
     */
    public int getNumberOfChanges() {
        return refactoring.getNumberOfChanges();
    }

    /**
     * Checks if there would be or were changes.
     * 
     * @return true if the refactorings can generalize the selection.
     */
    public boolean wasApplicable() {
        if (checkOnly) {
            return !refactoringStatus.hasFatalError();
        } else {
            return refactoring.getNumberOfChanges() > 0;
        }
    }

    /**
     * Retrieves the merged status of the operation.
     * 
     * @return a RefactoringStatus
     */
    public RefactoringStatus getRefactoringStatus() {
        return refactoringStatus;
    }
    
    private static SelectedElement getSelectedElement(ICompilationUnit icu, int typeStartPosition, int typeLength) {
        ASTParser parser = ASTParser.newParser(AST.JLS3);
        parser.setSource(icu);
        parser.setResolveBindings(true);
        CompilationUnit compilationUnit = (CompilationUnit) parser.createAST(null);
        SelectedElementVisitor vis = new SelectedElementVisitor(typeStartPosition, typeLength);
        compilationUnit.accept(vis);

        SelectedElement declarationElement = vis.getSelectedElement();
        if (declarationElement != null) {
            declarationElement.setCompilationUnit(icu);
        }
        return declarationElement;
    }
    

}
