/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3;

import java.io.IOException;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.WorkingCopyOwner;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TypeEnvironment;
import org.eclipse.ltk.core.refactoring.Change;
import org.eclipse.ltk.core.refactoring.Refactoring;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.constraintModel.creation.Collector;
import org.intoJ.inferType3.internal.constraintModel.util.InferTypeGmlHandler;
import org.intoJ.inferType3.internal.constraintModel.util.SelectedElement;
import org.intoJ.inferType3.internal.manipulation.Manipulation;
import org.intoJ.inferType3.internal.typeCalc.AccessSet;
import org.intoJ.inferType3.internal.typeCalc.AccessSetCalculation;
import org.intoJ.inferType3.internal.typeCalc.TypeCalculation;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.BindingWrapper;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.ExistingTCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TypeHierarchy;
import org.intoJ.inferType3.internal.util.Cache;
import org.intoJ.inferType3.internal.util.GmlWriter;
import org.intoJ.inferType3.internal.util.InternalCalculationException;

/**
 * The Infer Type Refactoring. 
 * 
 * This class dispatches all the work for Infer Type. {@link #checkInitialConditions(IProgressMonitor)} builds 
 * the constraint model and calculates the access sets. {@link #checkFinalConditions(IProgressMonitor)} 
 * calculates the new types and generates the necessary changes. They will be applied
 * by the framework. In addition, this class is the interface to all necessary data 
 * for the UI.
 * 
 * @author Hannes Kegel
 *
 */
public final class InferTypeRefactoring extends Refactoring {

    public static final String MSG_INTERNAL_ERROR = "Internal error";
    public static final String MSG_CREATING_CHANGES = "Creating changes";
    public static final String MSG_CALCULATING_ACCESS_SETS = "Calculating access sets";

    public static final int ERROR_CODE_TYPE_ARGUMENT_MODIFICATION_NECESSARY = 1;

    private final WorkingCopyOwner owner = new WorkingCopyOwner() { };

    private SelectedElement selectedElement;
    private TypeHierarchy typeHierarchy;

    private InferTypeConstraintsModel model;
    private Manipulation manipulation;
        
    private RefactoringStatus finalStatus;
    
    private boolean clearCache;

    private boolean noAbstractClasses = false;
    private boolean changeTypeArguments = true;
    private boolean redeclareDEs = true;
    
    private String newTypeName;
    private Iterable<BindingWrapper> checkedMembers;
    private Iterable<ExistingTCType> checkedInterfaces;
    
    private String newPackageName;
    private boolean ignoreSuperfluousMethods = false;
    
    /**
     * Creates a new Infer Type processor. 
     * 
     * The cache will be cleared.
     * 
     * @param declarationElement the selected declaration element
     */
    public InferTypeRefactoring(final SelectedElement declarationElement) {
        this(declarationElement, true);
    }

    /**
     * Creates a new Infer Type processor. 
     * 
     * @param declarationElement the selected declaration element
     * @param clearCache if the cache should be cleared
     */
    public InferTypeRefactoring(final SelectedElement declarationElement, boolean clearCache) {
        this.selectedElement = declarationElement;
        this.clearCache = clearCache;
        if (clearCache) {
            Cache.clearAll();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void finalize() throws Throwable {
        resetWorkingCopies();
    }

    /**
     * Resets all working copies.
     */
    private void resetWorkingCopies() {
        final ICompilationUnit[] units = JavaCore.getWorkingCopies(owner);
        for (final ICompilationUnit unit : units) {
            try {
                unit.discardWorkingCopy();
            } catch (Exception exception) {
                // Do nothing
            }
        }
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.ltk.core.refactoring.Refactoring#checkFinalConditions(org.eclipse.core.runtime.IProgressMonitor)
     */
    @Override
    public final RefactoringStatus checkFinalConditions(final IProgressMonitor monitor) throws CoreException, OperationCanceledException {
        Assert.isNotNull(monitor);
        if (finalStatus != null) {
            return finalStatus;
        }
        finalStatus = new RefactoringStatus();

        try {
            monitor.beginTask("", 100);
            monitor.setTaskName(InferTypeRefactoring.MSG_CREATING_CHANGES);

            // calculate which types should be used
            TypeCalculation typeCalculation = new TypeCalculation(model, typeHierarchy, noAbstractClasses, changeTypeArguments);            
            typeCalculation.determineTypes(newTypeName, checkedMembers, checkedInterfaces, redeclareDEs, finalStatus);
            monitor.worked(15);
                        
            if (!finalStatus.hasFatalError()) {
                // create all changes necessary to introduce the new types
                // do this here to be able to report errors with the refactoring status
                manipulation = new Manipulation(model, typeHierarchy, selectedElement);
                manipulation.calculateChanges(new SubProgressMonitor(monitor, 85), finalStatus, redeclareDEs, getValidationContext());
            }
        } catch (InternalCalculationException e) {
            finalStatus.addFatalError(e.getMessage());
        } finally {
            monitor.done();
        }

        return finalStatus;
    }
       
    /* (non-Javadoc)
     * @see org.eclipse.ltk.core.refactoring.Refactoring#checkInitialConditions(org.eclipse.core.runtime.IProgressMonitor)
     */
    @Override
    public final RefactoringStatus checkInitialConditions(final IProgressMonitor monitor) throws CoreException, OperationCanceledException {
        Assert.isNotNull(monitor);
        final RefactoringStatus status = new RefactoringStatus();
        try {
            TypeEnvironment environment = new TypeEnvironment();
            model = new InferTypeConstraintsModel(environment, selectedElement);
            
            // build hierarchy of refactored type
            final IType declaredType = (IType) selectedElement.getDeclaredType().getJavaElement();
            if (declaredType.getFullyQualifiedName().equals("java.lang.Object")
                    || declaredType.getFullyQualifiedName().equals("java.lang.String")) 
            {
                status.addFatalError("Declared type is already minimal");
            } else {
                typeHierarchy = new TypeHierarchy(declaredType, false, false);
    
                // collect affected files and build constraints
                                
                Collector collector = new Collector(owner, model, selectedElement.getBinding().getJavaElement().getJavaProject(), typeHierarchy, false, false, false, false, false, false);
                collector.run(monitor, status);
                
                // solve model
                AccessSetCalculation.calculate(model, typeHierarchy, status, ignoreSuperfluousMethods);
                newTypeName = typeHierarchy.getNewTypeName(model.getSelectedConstraintVariable(), this.newPackageName);
            }
            
        } catch (InternalCalculationException e) {
            status.addFatalError(e.getMessage());
        } finally {
            if (clearCache) {
                Cache.clearAll();
            }
        }
        return status;
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.ltk.core.refactoring.Refactoring#createChange(org.eclipse.core.runtime.IProgressMonitor)
     */
    @Override
    public final Change createChange(final IProgressMonitor monitor) throws CoreException, OperationCanceledException {
        Assert.isNotNull(monitor);
        try {
            monitor.beginTask("", 1);
            monitor.setTaskName("Modifing files");
            return manipulation.getAllChanges();
        } finally {
            monitor.done();
        }
    }

    /**
     * Sets whether no abstract classes should be created in case 3. If true, 
     * two interfaces will usually be created instead.
     * 
     * @param selection true if no abstract classes should be created
     */
    public void setNoAbstractClasses(boolean selection) {
        this.noAbstractClasses = selection;
    }

    /**
     * Returns whether no abstract classes should be created in case 3. If true, 
     * two interfaces will usually be created instead.
     * 
     * @return true if interfaces should be created instead
     */
    public boolean isNoAbstractClasses() {
        return noAbstractClasses;
    }
    

    /**
     * Returns whether type arguments should be changed if necessary.
     * If false, no changes will be performed if type arguments would be affected.
     * 
     * @return true if type argument changes are allowed
     */
    public boolean isChangeTypeArguments() {
        return changeTypeArguments;
    }

    /**
     * Sets whether type arguments should be changed if necessary.
     * If false, no changes will be performed if type arguments would be affected.
     * 
     * @param selection true if type argument changes are allowed
     */
    public void setChangeTypeArguments(boolean selection) {
        this.changeTypeArguments = selection;
    }    

    /**
     * Retrieves the access set of the selected declaration element. Call after
     * checkInitialConditions has run. 
     * 
     * @return the access set of the selected decl element.
     * @throws JavaModelException
     */
    public AccessSet getAccessSet() throws JavaModelException {
        return InferTypeConstraintsModel.getAccessSet(model.getSelectedConstraintVariable());
    }

    /**
     * Retrieves the type hierarchy of the selected declaration element. Call after
     * checkInitialConditions has run. 
     * 
     * @return the type hierarchy
     */
    public TypeHierarchy getTypeHierarchy() {
        return typeHierarchy;
    }

    /**
     * Used to set the additional public members as well as complete interfaces 
     * that will be included in the initial new type.
     * 
     * @param checkedMembers the additional methods.
     * @param checkedInterfaces the interfaces that the new type should be a sub type of (at least).
     */
    public void setCheckedMembers(Iterable<BindingWrapper> checkedMembers, Iterable<ExistingTCType> checkedInterfaces) {
        this.checkedMembers = checkedMembers;
        this.checkedInterfaces = checkedInterfaces;
    }

    /**
     * Gets the new type name of the initially created type.
     * Returns the default name if setNewTypeName has not been called.
     * 
     * @return the fully qualified type name 
     */
    public String getNewTypeName() {
        return newTypeName;
    }

    /**
     * Sets a custom new type name for the initially created type.
     *  
     * @param name the fully qualified type name
     */
    public void setNewTypeName(String name) {
        newTypeName = name;
    }
    
    /**
     * Checks if the inferred type should be used for the selected declaration element.
     * 
     * @return if true, the selected DE and all other necessary DEs will be redeclared
     * with the new type. 
     */
    public boolean isRedeclareDEs() {
        return redeclareDEs;
    }

    /**
     * Determines if the inferred type should be used for the selected declaration element.
     * 
     * @param redeclareDEs if true, the selected DE and all other necessary DEs will be redeclared
     * with the new type. 
     */
    public void setRedeclareDEs(boolean redeclareDEs) {
        this.redeclareDEs = redeclareDEs;
    }

    /**
     * If an existing type can be used as minimal type, returns the existing type name.
     * Otherwise returns null.
     * 
     * @return the fully qualified type name 
     * @throws JavaModelException
     */
    public String getFittingExistingTypeName() throws JavaModelException {
        TCType existingType = typeHierarchy.getFittingExistingType(getAccessSet());
        if (existingType != null) {
            return existingType.getTypeName();
        } else {
            return null;
        }
    }
    
    /**
     * Returns a description of other members of the original type or null if there are none.
     * 
     * @return the description or null
     */
    public String getOtherMemberDescription() {
        return typeHierarchy.getOtherMemberDescription();
    }
    

    /**
     * Determines if a new type will be created even if an existing type could be used.
     *  
     * @param forceNew true if a new type should be created in any case.
     */
    public void setForceNew(boolean forceNew) {
        typeHierarchy.setForceNewInitial(forceNew);
    }

    /**
     * Create a gml file from the current constraint graph. Call after
     * checkInitialConditions has run. More information about the file format can be found 
     * in {@link GmlWriter}
     * 
     * @param fileName the desired file name 
     * @param elementEdges if the graph should contain edges that show the relationship between element 
     * variables and its parametrized parent types 
     * @param onlyTraversed if only the part of the graph that will be traversed for IT should be exported
     * @throws IOException 
     * @throws JavaModelException 
     */
    public void exportGraph(String fileName, boolean elementEdges, boolean onlyTraversed) throws JavaModelException, IOException {
        InferTypeGmlHandler.writeGml(fileName, elementEdges, onlyTraversed, model, typeHierarchy);
    }

    /* (non-Javadoc)
     * @see org.eclipse.ltk.core.refactoring.Refactoring#getName()
     */
    public String getName() {
        return "Infer Type";
    }

    /**
     * Returns the number of changed type occurences. Used for statistics.
     * 
     * @return the number
     */
    public int getNumberOfChanges() {
        return manipulation == null ? 0 : manipulation.getNumberOfChanges();
    }

    /**
     * Sets the package name at which the new type will be added.
     * 
     * @param newPackageName the fully qualified package name
     */
	public void setNewPackageName(String newPackageName) {
		this.newPackageName = newPackageName;
	}
	
    /**
     * Returns the preferred package name at which the new type will be added.
     * 
     * @return the package name of the new type or <code>null</code> if the name
     * will be retrieved automatically
     */
	public String getNewPackageName() {
		return this.newPackageName;
	}
	
	/**
	 * Sets if the new type will be added even if it the old type is already
	 * minimal.
	 * 
	 * @param ignoreSuperfluousMethods <code>true</code> if a new type sould be
	 * added although the old is minimal, <code>false</code> if not.
	 */
	public void setIgnoreSuperfluousMethods(boolean ignoreSuperfluousMethods) {
		this.ignoreSuperfluousMethods = ignoreSuperfluousMethods;
	}
}