/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.constraintModel.creation;

import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.eclipse.core.runtime.Assert;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.AnonymousClassDeclaration;
import org.eclipse.jdt.core.dom.ArrayAccess;
import org.eclipse.jdt.core.dom.ArrayCreation;
import org.eclipse.jdt.core.dom.ArrayInitializer;
import org.eclipse.jdt.core.dom.ArrayType;
import org.eclipse.jdt.core.dom.Assignment;
import org.eclipse.jdt.core.dom.CastExpression;
import org.eclipse.jdt.core.dom.CatchClause;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.Comment;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.ConditionalExpression;
import org.eclipse.jdt.core.dom.ConstructorInvocation;
import org.eclipse.jdt.core.dom.EnhancedForStatement;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.FieldAccess;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.ImportDeclaration;
import org.eclipse.jdt.core.dom.InstanceofExpression;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.NullLiteral;
import org.eclipse.jdt.core.dom.PackageDeclaration;
import org.eclipse.jdt.core.dom.ParenthesizedExpression;
import org.eclipse.jdt.core.dom.QualifiedName;
import org.eclipse.jdt.core.dom.ReturnStatement;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.SuperConstructorInvocation;
import org.eclipse.jdt.core.dom.SuperFieldAccess;
import org.eclipse.jdt.core.dom.SuperMethodInvocation;
import org.eclipse.jdt.core.dom.ThisExpression;
import org.eclipse.jdt.core.dom.ThrowStatement;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.TypeLiteral;
import org.eclipse.jdt.core.dom.VariableDeclarationExpression;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.eclipse.jdt.internal.corext.SourceRange;
import org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.CompilationUnitRange;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.TypeVariable2;
import org.eclipse.jdt.internal.corext.refactoring.util.RefactoringASTParser;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.constraintModel.util.AnonWorkaround;
import org.intoJ.inferType3.internal.constraintModel.variables.ElementVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.GenericMethodTypeArgumentVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ImmutabilityReason;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.ExistingTCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TypeHierarchy;
import org.intoJ.inferType3.internal.util.BindingUtil;

/**
 * Type constraints creator to determine the necessary constraints to calculate
 * and introduce a minimal type.
 * <p>
 * The creator is an ASTVisitor that is called for all affected CUs by the
 * {@link Collector}. For all relevant language constructs (like assignments
 * and method invocations) it creates constraint variables for the involved
 * parts (like the variables, type declarations, parameters) and constraints
 * (subtype or equality) between them. The variables are bound to the AST for use
 * in other visitor call back methods.
 * <p>
 * <p>
 * The real creation of the variables and constraints is done by the 
 * {@link InferTypeConstraintsModel}. The model also checks that certain variables 
 * are unique, thus allowing to build up a connected graph.
 * </p>
 * <p>
 * Two of the most complicated tasks are sourced out into {@link MethodHandler} and
 * {@link TypeDeclElementHandler} for clarity.
 * </p>  
 *
 * @author Hannes Kegel
 *
 */
public class InferTypeConstraintsCreator extends HierarchicalASTVisitor {

    /** The constraint variable property */
    private static final String PROPERTY_CONSTRAINT_VARIABLE = "cv";

    /** The current method declarations being processed */
    private final Stack<MethodDeclaration> currentMethods = new Stack<MethodDeclaration>();

    /** The type constraint model to solve */
    private final InferTypeConstraintsModel model;
    
    /** the handler for method invocations and declarations */
    private MethodHandler methodHandler;

    /** the handler for parameterized supertypes */
    private TypeDeclElementHandler typeDeclElementHandler;

    /** The current types being parsed. Used for inner this-references */
    private final Stack<ITypeBinding> currentTypeStack = new Stack<ITypeBinding>();
    
    /** all parsed icus */
    private final Set<ICompilationUnit> parsedUnits = new HashSet<ICompilationUnit>();
    
    private boolean includeThisReceiver;
    private boolean includeThisReceiverForSubclassFields;
    private boolean includeSuperMethodInvocations;
	private boolean includeCasts;
	private boolean includeInstanceOf;
    private TypeHierarchy typeHierarchy;

    /**
     * The single constructor.
     * 
     * @param model the model to be filled
     * @param includeThisReceiverForSubclassFields2 
     * @param typeHierarchy 
     * @param status the current refactoring status used to report errors
     */
    public InferTypeConstraintsCreator(final InferTypeConstraintsModel model, boolean includeThisReceiver, boolean includeThisReceiverForSubclassFields, boolean includeSuperMethodInvocations, boolean includeCasts, boolean includeInstanceOf, TypeHierarchy typeHierarchy, RefactoringStatus status) {
		Assert.isNotNull(model);

        this.includeThisReceiver = includeThisReceiver;
		this.includeThisReceiverForSubclassFields = includeThisReceiverForSubclassFields;
		this.includeSuperMethodInvocations = includeSuperMethodInvocations;
		this.includeCasts = includeCasts;
		this.includeInstanceOf = includeInstanceOf;
        this.model = model;
        this.typeHierarchy = typeHierarchy;
        
        typeDeclElementHandler = new TypeDeclElementHandler();
        methodHandler = new MethodHandler(typeDeclElementHandler, model, status);
    }
    
    /**
     * Called when all units are parsed. Used primarily to match parametrized supertypes
     * with relevant method declarations and invocations.
     */
    public void endCreation() {
        typeDeclElementHandler.finish(model);
        model.endCreation();
    }
    
    /**
     * Stores a constraint variable for the given ASTNode. Also used by the MethodHandler. 
     * @param node the node
     * @param cv the constraint variable for the given node
     */
    static void setConstraintVariable(ASTNode node, ConstraintVariable2 cv) {
        if (node != null) {
            node.setProperty(PROPERTY_CONSTRAINT_VARIABLE, cv);
        }
    }

    /**
     * Retrieves the constraint variable for the given node.
     * 
     * @param node the node
     * @return the constraint variable or null if there is none stored
     */
    static ConstraintVariable2 getConstraintVariable(ASTNode node) {
        if (node == null) {
            return null;
        }
        ConstraintVariable2 cv = (ConstraintVariable2) node.getProperty(PROPERTY_CONSTRAINT_VARIABLE);
        return cv;
    }
    
    
    /* (non-Javadoc)
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#visit(org.eclipse.jdt.core.dom.CompilationUnit)
     */
    public boolean visit(CompilationUnit node) {
        // for the second pass in Collector, this could be called twice.
        ICompilationUnit icu = (ICompilationUnit)node.getJavaElement();
        if (parsedUnits.contains(icu)) {
            return false;
        } else {
            parsedUnits.add(icu);
            return true;            
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.ArrayAccess)
     */
    @Override
    public final void endVisit(final ArrayAccess node) {
        setConstraintVariable(node, getConstraintVariable(node.getArray()));
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.ArrayCreation)
     */
    @Override
    public final void endVisit(final ArrayCreation node) {
        final ConstraintVariable2 ancestor = getConstraintVariable(node.getType());
        setConstraintVariable(node, ancestor);
        final ArrayInitializer initializer = node.getInitializer();
        if (initializer != null) {
            final ConstraintVariable2 descendant = getConstraintVariable(initializer);
            if (descendant != null) {
                model.createSubtypeConstraint(descendant, ancestor);
            }
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.ArrayInitializer)
     */
    @Override
    public final void endVisit(final ArrayInitializer node) {
        final ITypeBinding binding = node.resolveTypeBinding();
        if (binding != null && binding.isArray()) {
            final ConstraintVariable2 ancestor = model.makeIndependentTypeVariable(binding.getElementType());
            setConstraintVariable(node, ancestor);
            Expression expression = null;
            ConstraintVariable2 descendant = null;
            final List expressions = node.expressions();
            for (int index = 0; index < expressions.size(); index++) {
                expression = (Expression) expressions.get(index);
                descendant = getConstraintVariable(expression);
                if (descendant != null) {
                    model.createSubtypeConstraint(descendant, ancestor);
                }
            }
        }
    }
    

    /* (non-Javadoc)
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.EnhancedForStatement)
     */
    @Override
    public void endVisit(EnhancedForStatement node) {
        Expression expression = node.getExpression();
        SingleVariableDeclaration parameter = node.getParameter();

        ConstraintVariable2 expressionCv = getConstraintVariable(expression);
        ConstraintVariable2 parameterCv = getConstraintVariable(parameter);
        ConstraintVariable2 nameCv = getConstraintVariable(parameter.getName());
        
        if (expressionCv != null && parameterCv != null && nameCv != null) {
            model.createEqualityConstraint(parameterCv, nameCv);

            ITypeBinding expressionBinding = expression.resolveTypeBinding();
            
            if (expressionBinding.isArray()) {
                model.createSubtypeConstraint(expressionCv, parameterCv);
            } else {
                ITypeBinding iterableBinding = BindingUtil.findTypeInHierarchy(expressionBinding.getTypeDeclaration(), "java.lang.Iterable");
                
                if (iterableBinding != null) {
                    String iterableTypeKey = iterableBinding.getTypeDeclaration().getTypeParameters()[0].getKey();
                    
                    ElementVariable elementVariable = InferTypeConstraintsModel.getElementVariables(expressionCv).get(iterableTypeKey);
                    if (elementVariable != null) {
                        model.createSubtypeConstraint(elementVariable, parameterCv);                        
                    }
                                        
                    ConstraintVariable2 iterableVar = model.makeImmutableTypeVariable(iterableBinding, expression, "used as Iterable in enhanced for-loop", null, ImmutabilityReason.OTHER);
                    model.createSubtypeConstraint(expressionCv, iterableVar);
                }
            }
        }
        
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.ArrayType)
     */
    @Override
    public final void endVisit(final ArrayType node) {
        Type elementType = node.getElementType();
        final ConstraintVariable2 variable = model.makeTypeVariable(elementType, true);
        if (variable != null) {
            setConstraintVariable(elementType, variable);
            setConstraintVariable(node, variable);
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.Assignment)
     */
    @Override
    public final void endVisit(final Assignment node) {
        final ConstraintVariable2 ancestor = getConstraintVariable(node.getLeftHandSide());
        final ConstraintVariable2 descendant = getConstraintVariable(node.getRightHandSide());
        setConstraintVariable(node, ancestor);
        if (ancestor != null && descendant != null) {
            model.createSubtypeConstraint(descendant, ancestor);
        }
    }

    
    @Override
	public void endVisit(InstanceofExpression node) {
    	if (includeInstanceOf) {
	    	ITypeBinding typeBinding = node.getRightOperand().resolveBinding();
	    	if (typeBinding != null) {
	    		typeBinding = typeBinding.getTypeDeclaration();
				// TODO: replace with more specific constraint 
	    		//model.makeImmutableTypeVariable(model.createTType(typeBinding), "Used in instanceof expression.", null, ImmutabilityReason.INSTANCE_OF);
	    	}
    	}
	    super.endVisit(node);
	}

	/*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.CastExpression)
     */
    @Override
    public final void endVisit(final CastExpression node) {
        final ConstraintVariable2 first = getConstraintVariable(node.getType());
        if (first != null) {
            setConstraintVariable(node, first);
            final ConstraintVariable2 second = getConstraintVariable(node.getExpression());
            if (second != null) {
                model.createElementEqualsConstraints(second, first, true, true);
                if (includeCasts) {
                	model.createSubtypeConstraint(first, second);
    	    		ITypeBinding binding = node.resolveTypeBinding();
    	    		if (binding != null) {
    					// TODO: replace with more specific constraint 
    	    			//model.makeImmutableTypeVariable(model.createTType(binding.getTypeDeclaration()), "Used in cast.", null, ImmutabilityReason.CAST);
    	    		}
                }
            }
            
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.CatchClause)
     */
    @Override
    public final void endVisit(final CatchClause node) {
        final SingleVariableDeclaration declaration = node.getException();
        if (declaration != null) {
            final ConstraintVariable2 descendant = getConstraintVariable(declaration);
            if (descendant != null) {
                final ITypeBinding binding = node.getAST().resolveWellKnownType("java.lang.Throwable"); //$NON-NLS-1$
                if (binding != null) {
                    final ConstraintVariable2 ancestor = model.makeImmutableTypeVariable(binding, null, "caught as exception", null, ImmutabilityReason.OTHER);
                    if (ancestor != null) {
                        model.createSubtypeConstraint(descendant, ancestor);
                    }
                }
            }
        }
    }

    /* (non-Javadoc)
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#visit(org.eclipse.jdt.core.dom.AbstractTypeDeclaration)
     */
    @Override
    public boolean visit(AbstractTypeDeclaration node) {
        currentTypeStack.push(node.resolveBinding());
        return super.visit(node);
    }

    /* (non-Javadoc)
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#visit(org.eclipse.jdt.core.dom.AnonymousClassDeclaration)
     */
    @Override
    public boolean visit(AnonymousClassDeclaration node) {
        currentTypeStack.push(node.resolveBinding());
        return super.visit(node);
    }
    

    /* (non-Javadoc)
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.AbstractTypeDeclaration)
     */
    @Override
    public void endVisit(AbstractTypeDeclaration node) {
        currentTypeStack.pop();
        super.endVisit(node);
    }

    /* (non-Javadoc)
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.AnonymousClassDeclaration)
     */
    @Override
    public void endVisit(AnonymousClassDeclaration node) {
        currentTypeStack.pop();
        super.endVisit(node);
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.ClassInstanceCreation)
     */
    @Override
    public final void endVisit(final ClassInstanceCreation node) {
        Type type = node.getType();

        IMethodBinding binding = null;
        if (node.getAnonymousClassDeclaration() != null) {            
            binding = AnonWorkaround.getSuperConstructorBinding(node.resolveConstructorBinding());
            if (type.isParameterizedType()) {
                ITypeBinding declaringType = node.resolveTypeBinding();
                TypeVariable2 var = model.makeTypeVariable(type, false);
                typeDeclElementHandler.registerTypeVariable(model.createTType(declaringType), var);
                
            }
        } else {
            binding = node.resolveConstructorBinding();
        }
        
        if (binding != null) {            
            ConstraintVariable2 variable = model.makeTypeVariable(type, true);
            if (variable != null) {
                setConstraintVariable(node, variable);
            }            
            methodHandler.handleInvocationParameters(node.arguments(), binding, node, null, currentTypeStack, includeThisReceiver);
        }
    }
    

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.ConditionalExpression)
     */
    @Override
    public final void endVisit(final ConditionalExpression node) {
        ConstraintVariable2 thenVariable = null;
        ConstraintVariable2 elseVariable = null;
        final Expression thenExpression = node.getThenExpression();
        if (thenExpression != null) {
            thenVariable = getConstraintVariable(thenExpression);
        }
        final Expression elseExpression = node.getElseExpression();
        if (elseExpression != null) {
            elseVariable = getConstraintVariable(elseExpression);
        }
        ITypeBinding binding = node.resolveTypeBinding();
        if (binding != null) {
            if (binding.isArray()) {
                binding = binding.getElementType();
            }
            final ConstraintVariable2 ancestor = model.makeIndependentTypeVariable(binding);
            if (ancestor != null) {
                setConstraintVariable(node, ancestor);
                if (thenVariable != null) {
                    model.createSubtypeConstraint(thenVariable, ancestor);
                }
                if (elseVariable != null) {
                    model.createSubtypeConstraint(elseVariable, ancestor);
                }
            }
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.ConstructorInvocation)
     */
    @Override
    public final void endVisit(final ConstructorInvocation node) {
        final IMethodBinding binding = node.resolveConstructorBinding();
        if (binding != null) {
            methodHandler.handleInvocationParameters(node.arguments(), binding, null, null, currentTypeStack, includeThisReceiver);
        }
        
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.FieldAccess)
     */
    @Override
    public final void endVisit(final FieldAccess node) {
        final IVariableBinding binding = node.resolveFieldBinding();
        if (binding != null) {
            handleVariableUsage(binding, node.getExpression(), node);
        }
    }
        

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.FieldDeclaration)
     */
    @Override
    public final void endVisit(final FieldDeclaration node) {
        handleVariableDeclaration(node.fragments(), node.getType(), node);
    }
    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.MethodDeclaration)
     */
    @Override
    public final void endVisit(final MethodDeclaration node) {
        currentMethods.pop();
        methodHandler.handleMethodDeclaration(node);
    }

    /* (non-Javadoc)
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.ThrowStatement)
     */
    @Override
    public void endVisit(ThrowStatement node) {
        final ConstraintVariable2 expressionCv = getConstraintVariable(node.getExpression());
        
        final ITypeBinding throwable = node.getAST().resolveWellKnownType("java.lang.Throwable"); //$NON-NLS-1$
        if (expressionCv != null && throwable != null) {
            ConstraintVariable2 ancestor = model.makeImmutableTypeVariable(throwable, null, "thrown as exception", null, ImmutabilityReason.OTHER);
            if (ancestor != null) {
                model.createSubtypeConstraint(expressionCv, ancestor);
            }
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.MethodInvocation)
     */
    @Override
    public final void endVisit(final MethodInvocation node) {
        //LogUtil.debugOut2(this, "-------------------------------------------------------------------");
        //LogUtil.debugOut2(this, "START VISIT " + binding + " on " + node.getExpression());
        methodHandler.handleInvocation(node, currentTypeStack, includeThisReceiver);
        //LogUtil.debugOut2(this, "END VISIT " + binding + " on " + node.getExpression());
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.NullLiteral)
     */
    @Override
    public final void endVisit(final NullLiteral node) {
        setConstraintVariable(node, model.makeImmutableTypeVariable(node.resolveTypeBinding(), null, "null literal", null, ImmutabilityReason.OTHER));
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.ParenthesizedExpression)
     */
    @Override
    public final void endVisit(final ParenthesizedExpression node) {
        setConstraintVariable(node, getConstraintVariable(node.getExpression()));
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.QualifiedName)
     */
    @Override
    public final void endVisit(final QualifiedName node) {
        final ASTNode parent = node.getParent();
        final Name qualifier = node.getQualifier();
        IBinding binding = qualifier.resolveBinding();
        if (binding instanceof ITypeBinding) {
            final ConstraintVariable2 variable = model.makeTypeVariable((ITypeBinding) binding, new CompilationUnitRange(RefactoringASTParser.getCompilationUnit(node), new SourceRange(qualifier
                    .getStartPosition(), qualifier.getLength())));
            if (variable != null) {
                setConstraintVariable(qualifier, variable);
            }
        }
        binding = node.getName().resolveBinding();
        if (binding instanceof IVariableBinding && !(parent instanceof ImportDeclaration)) {
            handleVariableUsage((IVariableBinding) binding, qualifier, node);            
        }
        else if (binding instanceof ITypeBinding && parent instanceof MethodDeclaration) {
            handleExceptionType((ITypeBinding) binding, node);
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.ReturnStatement)
     */
    @Override
    public final void endVisit(final ReturnStatement node) {
        final Expression expression = node.getExpression();
        if (expression != null) {
            final ConstraintVariable2 descendant = getConstraintVariable(expression);
            if (descendant != null) {
                final MethodDeclaration declaration = currentMethods.peek();
                if (declaration != null) {
                    final IMethodBinding binding = declaration.resolveBinding();
                    if (binding != null) {
                        final ConstraintVariable2 ancestor = model.makeReturnTypeVariable(binding, false);
                        if (ancestor != null) {
                            setConstraintVariable(node, ancestor);
                            model.createSubtypeConstraint(descendant, ancestor);
                        }
                    }
                }
            }
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.SimpleName)
     */
    @Override
    public final void endVisit(final SimpleName node) {
        final ASTNode parent = node.getParent();
        if (!(parent instanceof ImportDeclaration) && !(parent instanceof PackageDeclaration) && !(parent instanceof AbstractTypeDeclaration)) {
            final IBinding binding = node.resolveBinding();
            if (binding instanceof IVariableBinding && !(parent instanceof MethodDeclaration)) {
                handleVariableUsage((IVariableBinding) binding, null, node);
            }
            else if (binding instanceof ITypeBinding && parent instanceof MethodDeclaration) {
                handleExceptionType((ITypeBinding) binding, node);
            }
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.SingleVariableDeclaration)
     */
    @Override
    public final void endVisit(final SingleVariableDeclaration node) {
        final ConstraintVariable2 ancestor = getConstraintVariable(node.getType());
        if (ancestor != null) {
            setConstraintVariable(node, ancestor);
            final Expression expression = node.getInitializer();
            if (expression != null) {
                final ConstraintVariable2 descendant = getConstraintVariable(expression);
                if (descendant != null) {
                    model.createSubtypeConstraint(descendant, ancestor);
                }
            }
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.SuperConstructorInvocation)
     */
    @Override
    public final void endVisit(final SuperConstructorInvocation node) {
        final IMethodBinding binding = node.resolveConstructorBinding();
        if (binding != null) {
            methodHandler.handleInvocationParameters(node.arguments(), binding, node.getExpression(), null, currentTypeStack, includeThisReceiver);
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.SuperFieldAccess)
     */
    @Override
    public final void endVisit(final SuperFieldAccess node) {
        final Name name = node.getName();
        final IBinding binding = name.resolveBinding();
        if (binding instanceof IVariableBinding) {
            handleVariableUsage((IVariableBinding) binding, null, node);
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.SuperMethodInvocation)
     */
    @Override
    public final void endVisit(final SuperMethodInvocation node) {
        final IMethodBinding superBinding = node.resolveMethodBinding();
        if (superBinding != null) {
            methodHandler.handleInvocationParameters(node.arguments(), superBinding, null, null, currentTypeStack, includeThisReceiver);
            
            ConstraintVariable2 retTypeVar = methodHandler.handleInvocationReturn(null, superBinding, null);
            if (retTypeVar != null) {
                setConstraintVariable(node, retTypeVar);
            }
            
            if (includeSuperMethodInvocations) {
            	ITypeBinding thisBinding = currentTypeStack.peek();
            	if (node.getQualifier() != null) {
            		thisBinding = node.getQualifier().resolveTypeBinding();
            	}
            	ConstraintVariable2 cv = model.makeSuperVariable(thisBinding);
            	methodHandler.handleDeclarationRequirements(superBinding, cv);
            }
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.ThisExpression)
     */
    @Override
    public final void endVisit(final ThisExpression node) {
        final ITypeBinding binding = node.resolveTypeBinding();
        if (binding != null) {
            setConstraintVariable(node, model.makeThisVariable(binding));
        }
    }
       
    @Override
    public final void endVisit(final Type node) {
        final ASTNode parent = node.getParent();
        if (!(parent instanceof TypeLiteral)) {
            if (parent instanceof AbstractTypeDeclaration) {
                if (node.isParameterizedType()) {
                    ITypeBinding declaringType = ((AbstractTypeDeclaration)parent).resolveBinding();
                    TypeVariable2 var = model.makeTypeVariable(node, false);
                    typeDeclElementHandler.registerTypeVariable(model.createTType(declaringType), var);
                    setConstraintVariable(node, var);    
                }
                //TODO: IT4: support mutable bounds
/*            } else if (parent instanceof TypeParameter && parent.getParent() instanceof AbstractTypeDeclaration) {
                TypeParameter typeParameter = (TypeParameter)parent;                
                TypeVariable typeVariable = (TypeVariable)fModel.createTType(typeParameter.resolveBinding());
                TType[] bounds = typeVariable.getBounds();
                if (bounds.length == 1) {
                    ASTNode boundNode = (ASTNode)typeParameter.typeBounds().get(0);
                    if (boundNode instanceof Type) {
                        BoundVariable boundVariable = fModel.makeBoundVariable(typeVariable, 0, bounds[0], true);
                        ConstraintVariable2 boundDeclTypeVar = fModel.makeTypeVariable((Type)boundNode);
                        boundNode.setProperty(PROPERTY_CONSTRAINT_VARIABLE, boundDeclTypeVar);
                        fModel.createEqualityConstraint(boundVariable, boundDeclTypeVar);
                    }
                }
                System.out.println("bound type: " + node + ", var key: " + typeParameter.resolveBinding().getKey());
                */                
            } else if (!(parent instanceof ClassInstanceCreation)
                    && (!(parent instanceof InstanceofExpression))) {
                ConstraintVariable2 typeVariable = model.makeTypeVariable(node, true);
                ITypeBinding typeBinding = node.resolveBinding();
                if (typeBinding != null && !typeBinding.isFromSource()) {
                    ConstraintVariable2 immutableCv = model.makeImmutableTypeVariable(typeBinding, null, "used as non-source type " + typeBinding.getQualifiedName(), null, ImmutabilityReason.OTHER);
                    model.createSubtypeConstraint(typeVariable, immutableCv);
                }
                
                setConstraintVariable(node, typeVariable);
            }
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.VariableDeclarationExpression)
     */
    @Override
    public final void endVisit(final VariableDeclarationExpression node) {
        handleVariableDeclaration(node.fragments(), node.getType(), node);
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.VariableDeclarationFragment)
     */
    @Override
    public final void endVisit(final VariableDeclarationFragment node) {
        final Expression initializer = node.getInitializer();
        if (initializer != null) {
            setConstraintVariable(node, getConstraintVariable(initializer));
        }
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#endVisit(org.eclipse.jdt.core.dom.VariableDeclarationStatement)
     */
    @Override
    public final void endVisit(final VariableDeclarationStatement node) {
        handleVariableDeclaration(node.fragments(), node.getType(), node);
    }
           
    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#visit(org.eclipse.jdt.core.dom.Comment)
     */
    @Override
    public final boolean visit(final Comment node) {
        return false;
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#visit(org.eclipse.jdt.core.dom.ImportDeclaration)
     */
    @Override
    public final boolean visit(final ImportDeclaration node) {
        return false;
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#visit(org.eclipse.jdt.core.dom.MethodDeclaration)
     */
    @Override
    public final boolean visit(final MethodDeclaration node) {
        currentMethods.push(node);
        return super.visit(node);
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#visit(org.eclipse.jdt.core.dom.PackageDeclaration)
     */
    @Override
    public final boolean visit(final PackageDeclaration node) {
        return false;
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#visit(org.eclipse.jdt.core.dom.ThisExpression)
     */
    @Override
    public final boolean visit(final ThisExpression node) {
        return false;
    }

    /*
     * @see org.eclipse.jdt.internal.corext.dom.HierarchicalASTVisitor#visit(org.eclipse.jdt.core.dom.Type)
     */
    @Override
    public final boolean visit(final Type node) {
        return false;
    }
    


    /**
     * Handle a thrown exception.
     * 
     * @param binding the type binding of the thrown exception
     * @param node the exception name node
     */
    private void handleExceptionType(final ITypeBinding binding, final Name node) {
        final ConstraintVariable2 variable = model.makeExceptionVariable(node);
        if (variable != null) {
            setConstraintVariable(node, variable);
        }
    }
    
    private ConstraintVariable2 getFieldVariable(final Expression expression, final IVariableBinding binding) {
        IVariableBinding variableDeclaration = binding.getDeclaringClass() == null || binding.getDeclaringClass().isRawType() ? binding : binding.getVariableDeclaration();
        if (expression != null && MethodHandler.containsWildcardsOrTypeVars(variableDeclaration.getType())) {
            //if this is a type var or wildcard field, it is handled like a getter return type.
            ITypeBinding typeBinding = variableDeclaration.getType();
            if (typeBinding.isArray()) {
                typeBinding = typeBinding.getElementType();
            }
            ConstraintVariable2 expressionCv = InferTypeConstraintsCreator.getConstraintVariable(expression);
            if (typeBinding.isTypeVariable()) {
                return model.makeElementVariable(expressionCv, typeBinding, model.createTType(binding.getType()), true); 
            } else if (typeBinding.isParameterizedType()) {
                TType declaredReturnTType= model.createTType(typeBinding);
                ConstraintVariable2 variable = model.makeIndependentFieldVariable(binding);
                Map<String, GenericMethodTypeArgumentVariable> genMethodTypeVars = Collections.emptyMap();
                model.createTypeVariablesEqualityConstraints(expressionCv, genMethodTypeVars, variable, declaredReturnTType, model.createTType(binding.getType()));
                return variable;
            }
        }
        return model.makeVariableVariable(variableDeclaration);            
    }        
    

    /**
     * Handle field access.
     * 
     * @param binding the variable binding
     * @param qualifier the qualifier expression, or <code>null</code>
     * @param access the access expression
     */
    private void handleVariableUsage(final IVariableBinding binding, final Expression qualifier, final Expression access) {
        ConstraintVariable2 variableVariable = getFieldVariable(qualifier, binding);
        
        setConstraintVariable(access, variableVariable);
        ConstraintVariable2 descendant = null;
        if (qualifier != null  && !(qualifier instanceof ThisExpression)) {
            descendant = getConstraintVariable(qualifier);
        } else if (includeThisReceiverForSubclassFields) {
        	ITypeBinding calledOnType = null;
            if (qualifier instanceof ThisExpression) {
                ThisExpression thisExpression = (ThisExpression)qualifier;
                calledOnType = thisExpression.resolveTypeBinding();
            } else {
                ITypeBinding declaringClass = binding.getDeclaringClass();

                for (int i=currentTypeStack.size() - 1; i >=0; i--) {
                    ITypeBinding currentType = currentTypeStack.get(i); 
                    if (BindingUtil.isFirstSubclassOrEqual(currentType, declaringClass)) {
                    	calledOnType = currentType;
                        break;
                    }
                }
            }
            if (calledOnType != null && isSubType(calledOnType)) {
            	descendant = model.makeThisVariable(calledOnType);
            }
        }
        if (descendant != null) {
            if (variableVariable != null) {
                InferTypeConstraintsModel.setSimpleName(descendant, variableVariable);
            }

            final ITypeBinding declaringClass = binding.getDeclaringClass();
            if (declaringClass != null) {
                // array.length does not have a declaring class
                final ConstraintVariable2 ancestor = model.makeImmutableDeclaringTypeVariable(declaringClass, "access to field " + binding.getDeclaringClass().getQualifiedName() + "." + binding.getName(), binding.getKey(), ImmutabilityReason.FIELD_ACCESS);
                if (ancestor != null) {
                    if (descendant != null) {
                        InferTypeConstraintsModel.setSimpleName(descendant, variableVariable);
                        model.createSubtypeConstraint(descendant, ancestor);
                    }
                }
            }
        }
        
        
        ITypeBinding type = binding.getType();
        if (variableVariable != null && binding.isField() && !binding.getDeclaringClass().isFromSource()) {
            String description = "used in non-source type " + binding.getDeclaringClass().getQualifiedName();
            model.createSubtypeConstraint(variableVariable, model.makeImmutableTypeVariable(model.createTType(type), description, null, ImmutabilityReason.OTHER));
            model.makeElementsImmutable(variableVariable, description);
        }
                
    }

    private boolean isSubType(ITypeBinding possibleSub) {
		if (possibleSub.getKey().equals(typeHierarchy.getOriginalSelectedType().getBaseTypeBinding().getKey())) {
			return true;
		}
    	for (ExistingTCType existingTCType : typeHierarchy.getSubTypes()) {
			if (possibleSub.getKey().equals(existingTCType.getBaseTypeBinding().getKey())) {
				return true;
			}
		}
		return false;
	}

	/**
     * Handle a variable declaration fragment list.
     * 
     * @param fragments the fragments (element type: <code>VariableDeclarationFragment</code>)
     * @param type the type of the fragments
     * @param parent the parent of the fragment list
     */
    private void handleVariableDeclaration(final List fragments, final Type type, final ASTNode parent) {
        final ConstraintVariable2 ancestor = getConstraintVariable(type);
        if (ancestor != null) {
            for (int index = 0; index < fragments.size(); index++) {
                VariableDeclarationFragment fragment = (VariableDeclarationFragment) fragments.get(index);
                ConstraintVariable2 descendant = getConstraintVariable(fragment);
                if (descendant != null) {
                    model.createSubtypeConstraint(descendant, ancestor);
                }
                IVariableBinding binding = fragment.resolveBinding();
                if (binding != null) {
                    descendant = model.makeVariableVariable(binding);
                    if (descendant != null) {
                        model.createEqualityConstraint(ancestor, descendant);
                    }
                }
            }
            setConstraintVariable(parent, ancestor);
        }
    }

}
