/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.constraintModel.variables;

import java.util.EnumSet;

import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;

/**
 * An immutable type variable represents a type that cannot be changed. This can be
 * a type of a field access, an external type etc. The variable also holds a description
 * why this type is immutable. This description will be presented by the UI if this variable
 * is traversed starting from the selection.
 * 
 * @author Hannes Kegel
 *
 */
public class ImmutableTypeVariableWithDescr extends ConstraintVariable2 implements BindingKeyProvider {
	
    private final String description;
    private final String bindingKey;
    private final EnumSet<ImmutabilityReason> reason;

    /**
     * The single constructor.
     * 
     * @param type the type that cannot be changed
     * @param description an explanation why it cannot be changed (field access, external etc.)
     */
    public ImmutableTypeVariableWithDescr(TType type, String description, String bindingKey, ImmutabilityReason reason) {
        super(type);
        this.description = description;
		this.bindingKey = bindingKey;
		this.reason = EnumSet.of(reason);
    }

    public ImmutableTypeVariableWithDescr(TType type, String description, String bindingKey, EnumSet<ImmutabilityReason> reason) {
        super(type);
        this.description = description;
		this.bindingKey = bindingKey;
		this.reason = reason;
    }
    
    public String getBindingKey() {
		return bindingKey;
	}

	public EnumSet<ImmutabilityReason> getReason() {
		return reason;
	}

	/**
     * Returns why this type is immutable.
     * 
     * @return the description
     */
    public String getDescription() {
        return description;
    }

    
    @Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((bindingKey == null) ? 0 : bindingKey.hashCode());
		result = prime * result
				+ ((description == null) ? 0 : description.hashCode());
		result = prime * result
				+ ((fType == null) ? 0 : fType.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final ImmutableTypeVariableWithDescr other = (ImmutableTypeVariableWithDescr) obj;
		if (bindingKey == null) {
			if (other.bindingKey != null)
				return false;
		} else if (!bindingKey.equals(other.bindingKey))
			return false;
		if (description == null) {
			if (other.description != null)
				return false;
		} else if (!description.equals(other.description))
			return false;
		if (fType == null) {
			if (other.fType != null)
				return false;
		} else if (!fType.equals(other.fType))
			return false;
		return true;
	}

	/* (non-Javadoc)
     * @see org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2#toString()
     */
    @Override
    public String toString() {
        return getType().getPrettySignature();
    }
}
