/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.typeCalc.hierarchy;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeHierarchy;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.ArrayType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.CaptureType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.HierarchyType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TypeVariable;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.WildcardType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.intoJ.inferType3.InferTypePlugin;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.typeCalc.AccessSet;
import org.intoJ.inferType3.internal.typeCalc.CovarianceType;
import org.intoJ.inferType3.internal.typeCalc.TypeNameGenerator;
import org.intoJ.inferType3.internal.util.ConversionUtil;
import org.intoJ.inferType3.internal.util.InternalCalculationException;

/**
 * This class manages all {@link TCType} instances for the type hierarchy
 * of the selected type. It checks if existing types can be used for a
 * given constraint variable and creates new types if necessary. 
 * 
 * @author Hannes Kegel
 *
 */
public class TypeHierarchy {
    private TypeNameGenerator typeNameGenerator;
    
    private Map<IType, ExistingTCType> iTypeToUsedType = new HashMap<IType, ExistingTCType>();
    private HashSet<TCType> allTypes = new HashSet<TCType>();
    private Map<String, BindingWrapper> keyToWrapper = new HashMap<String, BindingWrapper>();
    private Map<BindingWrapper, BindingWrapper> allWrappers = new HashMap<BindingWrapper, BindingWrapper>();
    private Set<ExistingTCType> subTypes = new HashSet<ExistingTCType>();


    private ExistingTCType originalSelectedType;
    private ExistingTCType objectType;

    private boolean forceNewInitial = false;

    /**
     * The constructor builds up the hierarchy of existing types based on the argument.
     * 
     * @param initialType the selected type
     * @param includeSubtypes if all subtypes of the initial type should be included in the hierarchy   
     * @throws JavaModelException
     * @throws InternalCalculationException
     */
    public TypeHierarchy(IType initialType, boolean includeSubtypes, boolean allMethods) throws JavaModelException, InternalCalculationException {
        typeNameGenerator = new TypeNameGenerator(initialType.getJavaProject());
        originalSelectedType = buildHierarchy(initialType, includeSubtypes, allMethods);
    }

    /**
     * Get the minimal type for the selected element.
     * 
     * @param selectedVar the constraint variable of the selection
     * @param newTypeName the name if the type is new
     * @param status the refactoring status used to store errors
     * @return a minimal type, existing or new
     * @throws JavaModelException
     */
    public TCType getInitialMinimalType(ConstraintVariable2 selectedVar, String newTypeName, RefactoringStatus status) throws JavaModelException {
        TCType existingSubType = iTypeToUsedType.get(getHierarchyType(selectedVar.getType()).getJavaElementType());
        TCType ret = getType(selectedVar, existingSubType, newTypeName, forceNewInitial, status, CovarianceType.ARGUMENT_POSSIBLE_SUBTYPE);
        ret.setArgumentTType(selectedVar.getType());
        return ret;
    }

    /**
     * Get a matching type for a subtype relationship.
     * 
     * @param variable the variable for the type to be returned
     * @param subType the type that must be a subtype or equal to the returned type
     * @param status the refactoring status used to store errors
     * @return a type that can be used for the variable considering that subType must be compatible
     * @throws JavaModelException
     */
    public TCType getType(ConstraintVariable2 variable, TCType subType, RefactoringStatus status) throws JavaModelException {
        return getType(variable, subType, null, false, status, CovarianceType.NONE);
    }

    /**
     * Get the canonical name for a new type generated for this constraint variable.
     * 
     * @param variable the variable a new type should be created for
     * @param packageName the package name at which the new type sould be added.
     * <code>null</code> if it should be derived automatically
     * @return a canonical name described in {@link TypeNameGenerator}
     */
    public String getNewTypeName(ConstraintVariable2 variable, String packageName) {
        if (variable == null) {
            return "";
        }
        return typeNameGenerator.getName(variable, getHierarchyType(variable.getType()), packageName);
    }

    /**
     * Returns all new types.
     * 
     * @return all new types.
     */
    public Set<NewTCType> getNewTypes() {
        //LogUtil.debugOut(this, "getting new types");
        Set<NewTCType> ret = new HashSet<NewTCType>();
        for (TCType type : allTypes) {
            if (type instanceof NewTCType) {
                //LogUtil.debugOut(this, "found " + type.getTypeName());
                ret.add((NewTCType) type);
            }
        }
        return ret;
    }

    /**
     * Returns all existing types that have been modified.
     * 
     * @return all modified types
     */
    public Set<ExistingTCType> getModifiedTypes() {
        //LogUtil.debugOut(this, "getting modified types");
        Set<ExistingTCType> ret = new HashSet<ExistingTCType>();
        for (TCType type : allTypes) {
            if (type instanceof ExistingTCType && ((ExistingTCType) type).isModified()) {
                //LogUtil.debugOut(this, "found " + type.getTypeName());
                ret.add((ExistingTCType) type);
            }
        }
        return ret;
    }

    /**
     * Add superType as a new super type to changeType, if possible.
     * 
     * @param changeType the becoming sub type
     * @param superType the super type
     * @param status the refactoring status used to store errors
     * @return true if the new super type was added
     * @throws JavaModelException
     */
    public boolean addSupertype(TCType changeType, TCType superType, RefactoringStatus status) throws JavaModelException {
        if (superType.isObject()) {
            return true;
        } else if (changeType.equals(superType)) {
            return true;
        } else if (superType.isInterface()) {
            addInterface(changeType, superType, status);
            return true;
        } else if (!changeType.isInterface()) {
            changeType.removeInterface(superType);
            changeType.setSuperType(superType, status);
            return true;
        } else {
            status.addFatalError("tried to add class " + superType + " as super type of interface " + changeType);
            return false;
        }

    }

    /**
     * Return an {@link ExistingTCType} matching the given {@link TType} if
     * it exists in the hierarchy.
     * 
     * @param tType the TType
     * @return the existing type from the hierarchy or null if there is none
     */
    public ExistingTCType getType(TType tType) {
        HierarchyType hierarchyType = getHierarchyType(tType);
        if (hierarchyType == null) {
            return null;
        }
        return iTypeToUsedType.get(hierarchyType.getJavaElementType());
    }

    /**
     * Returns all existing and new types in this hierarchy.
     * 
     * @return all types
     */
    public Set<TCType> getAllTypes() {
        return allTypes;
    }

    /**
     * Returns the {@link BindingWrapper} for this method if the method
     * exists in this hierarchy.
     * 
     * @param bindingKey the key we search for
     * @return the binding wrapper or null if the method is not from this hierarchy
     */
    public BindingWrapper getMethod(String bindingKey) {
        return keyToWrapper.get(bindingKey);
    }

    /**
     * Update the existing {@link BindingWrapper} with the given ones.
     * 
     * @param newBindings the wrappers containing new {@link IMethodBinding} instances
     * @return true if anything was updated
     */
    public boolean updateWrappers(List<BindingWrapper> newBindings) {
        boolean changed = false;
        for (BindingWrapper wrapper : newBindings) {
            BindingWrapper oldWrapper = allWrappers.get(wrapper);
            if (oldWrapper != null && !oldWrapper.getMethodBinding().getKey().equals(wrapper.getMethodBinding().getKey())) {
                oldWrapper.setUpdatedMethodBinding(wrapper.getMethodBinding());
                changed = true;
            }
        }
        return changed;
    }

    /**
     * Check if the given type has a sub type that is an interface.
     * 
     * @param baseType the type to check
     * @return true if a subtype is an interace
     */
    public boolean hasInterfaceSubtype(TCType baseType) {
        for (TCType type : allTypes) {
            if (type.getAllSupertypes().contains(baseType) && type.isInterface()) {
                return true;
            }
        }
        return false;
    }
    
    public Set<ExistingTCType> getSubTypes() {
		return subTypes;
	}

	/**
     * Get the original type of the selection this hierarchy is based on.
     * 
     * @return the original existing type
     */
    public ExistingTCType getOriginalSelectedType() {
        return originalSelectedType;
    }

    /**
     * Returns a minimal existing type for the given access set if one can be found.
     * 
     * @return an existing type from the hierarchy or null
     */
    public TCType getFittingExistingType(AccessSet accessSet) {
        return tryExistingType(accessSet, originalSelectedType, true, CovarianceType.ARGUMENT_POSSIBLE_SUBTYPE);
    }

    /**
     * Returns a description of other members of the original type or null if there are none.
     * 
     * @return the description or null
     */
    public String getOtherMemberDescription() {
        return originalSelectedType.getOtherMemberDescription();
    }

    /**
     * Determine if a new type should be created even if there is a fitting existing one.
     * 
     * @param forceNewInitial if true a new type will be created in any case
     */
    public void setForceNewInitial(boolean forceNewInitial) {
        this.forceNewInitial = forceNewInitial;
    }
    
    private ExistingTCType buildHierarchy(IType initialType, boolean includeSubtypes, boolean allMethods) throws JavaModelException, InternalCalculationException {
        ConversionUtil.Result result = ConversionUtil.findBinding(initialType);

        ExistingTCType existingTCType = build((ITypeBinding)result.getBinding(), allMethods);
        objectType = build(result.getObjectBinding(), allMethods);
        
        if (includeSubtypes) {
        	ITypeHierarchy subHierarchy = initialType.newTypeHierarchy(null);
        	IType[] subtypes = subHierarchy.getAllSubtypes(initialType);
        	for (IType type : subtypes) {
        		try {
        			subTypes.add(buildHierarchy(type, false, allMethods));
        		} catch (InternalCalculationException e) {
        			InferTypePlugin.getDefault().getLog().log(new Status(IStatus.WARNING, InferTypePlugin.getDefault().getBundle().getSymbolicName(), 0, "Subtype cannot be parsed.", e));        			
        		}
			}
        }
        
        return existingTCType;
    }

    private ExistingTCType build(ITypeBinding binding, boolean allMethods) throws JavaModelException {
        if (binding == null) {
            return null;
        }
        ITypeBinding erasureBinding = binding.getErasure();
        
        ExistingTCType existingTCType = iTypeToUsedType.get(erasureBinding.getJavaElement());
        if (existingTCType == null) {
            Set<TCType> interfaces = new HashSet<TCType>();
            for (ITypeBinding currentIf : binding.getInterfaces()) {
                interfaces.add(build(currentIf, allMethods));
            }

            ITypeBinding superClass = binding.getSuperclass();
            existingTCType = new ExistingTCType(binding, build(superClass, allMethods), interfaces, allMethods);

            iTypeToUsedType.put((IType) erasureBinding.getJavaElement(), existingTCType);
            allTypes.add(existingTCType);
            for (BindingWrapper method : existingTCType.getDeclaredMethods()) {
                String key = method.getMethodBinding().getMethodDeclaration().getKey();
                keyToWrapper.put(key, method);
                allWrappers.put(method, method);
            }
        }
        return existingTCType;
    }

    private TCType getType(ConstraintVariable2 selectedVar, TCType existingSubType, String newTypeName, boolean forceNewNow, RefactoringStatus status, CovarianceType covarianceType) throws JavaModelException {
        AccessSet accessSet = InferTypeConstraintsModel.getAccessSet(selectedVar);
        assert (accessSet != null);

        if (!forceNewNow) {
            TCType existingType = tryExistingType(accessSet, existingSubType, newTypeName != null, covarianceType);
            if (existingType != null) {
                return existingType;
            }
        }

        NewTCType newTCType = new NewTCType(accessSet.getMembers(), this, getType(selectedVar.getType()));
        allTypes.add(newTCType);
        newTCType.setTypeName(newTypeName != null ? newTypeName : getNewTypeName(selectedVar, null));

        addInterface(existingSubType, newTCType, status);

        return newTCType;
    }
    
    private TCType tryExistingType(AccessSet accessSet, TCType currentType, boolean minimal, CovarianceType covarianceType) {
        if (accessSet.isEmpty() && !currentType.getAllMethods().isEmpty()) {
            return objectType;
        }
        if (currentType.isGenericType()) {
            if (checkType(accessSet, currentType, minimal, covarianceType)) {
                return currentType;
            } else {
                return null;
            }
        }
        Set<TCType> declaredSupertypes = currentType.getDeclaredSupertypes();
        for (TCType superType : declaredSupertypes) {
            TCType ret = tryExistingType(accessSet, superType, minimal, covarianceType);
            if (ret != null) {
                return ret;
            }
        }
        if (checkType(accessSet, currentType, minimal, covarianceType)) {
            return currentType;
        }
        return null;
    }

    private boolean checkType(AccessSet accessSet, TCType currentType, boolean minimal, CovarianceType covarianceType) {
        return accessSet.getMembers().equals(currentType.getAllMethods(), covarianceType) && 
                (!minimal || !currentType.hasOtherMembersToBeHidden())
                        && (minimal || currentType.canBeMovedInHierarchy());
    }

    private void addInterface(TCType changeType, TCType interf, RefactoringStatus status) throws JavaModelException {
        if (changeType.addInterface(interf, status)) {
            removeInterfaceFromSubtypes(changeType, interf);
        }
    }

    private void removeInterfaceFromSubtypes(TCType changeType, TCType interf) {
        for (TCType type : allTypes) {
            if (type.getAllSupertypes().contains(changeType)) {
                type.removeInterface(interf);
            }
        }
    }

    private HierarchyType getHierarchyType(TType tType) {
        if (tType instanceof CaptureType) {
            tType = ((CaptureType) tType).getWildcard();
        }
        if (tType instanceof WildcardType) {
            tType = ((WildcardType) tType).getBound();
        }
        if (tType instanceof ArrayType) {
            tType = ((ArrayType) tType).getElementType();
        }

        if (tType == null) {
            return null;
        } else  if (tType instanceof TypeVariable) {
            return (HierarchyType)((TypeVariable)tType).getErasure();
        } else if (tType instanceof HierarchyType) {
            return (HierarchyType) tType;
        }
        return null;
    }

    
}