/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.ui;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.Viewer;
import org.intoJ.inferType3.internal.typeCalc.AccessSet;
import org.intoJ.inferType3.internal.typeCalc.BindingSet;
import org.intoJ.inferType3.internal.typeCalc.CovarianceType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.BindingWrapper;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.ExistingTCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TypeHierarchy;

/**
 * The content provider for the tree in the Infer Type wizard.
 * 
 * @author Hannes Kegel
 *
 */
class ITContentProvider implements ITreeContentProvider {

    private AccessSet accessSet;
    private TypeHierarchy typeHierarchy;
    private Map<TCType, WrapperLeaf[]> childrenMap = new HashMap<TCType, WrapperLeaf[]>(); 
        
    /**
     * The single constructor
     * @param accessSet the access set of the selected element
     */
    public ITContentProvider(AccessSet accessSet) {
        this.accessSet = accessSet;
    }

    /* (non-Javadoc)
     * @see org.eclipse.jface.viewers.IStructuredContentProvider#getElements(java.lang.Object)
     */
    public Object[] getElements(Object element) {
            return getChildren(element);
    }

    /* (non-Javadoc)
     * @see org.eclipse.jface.viewers.IContentProvider#dispose()
     */
    public void dispose() {
        // Do nothing
    }

    /* (non-Javadoc)
     * @see org.eclipse.jface.viewers.IContentProvider#inputChanged(org.eclipse.jface.viewers.Viewer, java.lang.Object, java.lang.Object)
     */
    public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
        if (newInput instanceof TypeHierarchy) {
            typeHierarchy= (TypeHierarchy) newInput;
            
            childrenMap.clear();

            for (TCType type : typeHierarchy.getAllTypes()) {
                List<WrapperLeaf> children = new ArrayList<WrapperLeaf>();
                
                BindingSet bindingSet = ((ExistingTCType)type).getDeclaredMethods();
                for (BindingWrapper wrapper : bindingSet) {
                    if (!accessSet.getMembers().contains(wrapper, CovarianceType.RECEIVER_POSSIBLE_SUBTYPE)) {
                        children.add(new WrapperLeaf(type, wrapper));
                    }
                }
                
                childrenMap.put(type, children.toArray(new WrapperLeaf[children.size()]));
            }
        } else {
            typeHierarchy= null;
        }
    }

    /* (non-Javadoc)
     * @see org.eclipse.jface.viewers.ITreeContentProvider#getChildren(java.lang.Object)
     */
    public Object[] getChildren(Object parentElement) {
        if (parentElement instanceof TypeHierarchy) {
            List<Object> entries = new ArrayList<Object>();
            
            entries.addAll(accessSet.getMembers().getCollection());
            
            for (TCType type : typeHierarchy.getAllTypes()) {
                if (!type.isObject() && childrenMap.get(type).length > 0) {
                    entries.add(type);
                }                        
            }
            return entries.toArray();
        } else if (parentElement instanceof ExistingTCType) {
            return childrenMap.get(parentElement);
        } else {
            return new Object[0];
        }
    }

    /* (non-Javadoc)
     * @see org.eclipse.jface.viewers.ITreeContentProvider#getParent(java.lang.Object)
     */
    public Object getParent(Object element) {
        if (element instanceof WrapperLeaf) {
            return ((WrapperLeaf)element).getParent();
        } else if (element instanceof TypeHierarchy) {
            return null;
        } else {
            return typeHierarchy;
        }
    }

    /* (non-Javadoc)
     * @see org.eclipse.jface.viewers.ITreeContentProvider#hasChildren(java.lang.Object)
     */
    public boolean hasChildren(Object element) {
        return getChildren(element).length > 0;
    }
}
